package ActiveState::Milter;

use strict;

our $VERSION = '1.002';

my %command = (
    'O' => 'option',
    'D' => 'def_macro',
    'C' => 'connect',
    'H' => 'helo',

    'M' => 'mail_from',
    'R' => 'rcpt_to',
    'L' => 'header',
    'N' => 'eoh',
    'B' => 'body',
    'E' => 'eom',

    'A' => 'abort',
    'Q' => 'quit',
);

while (my($k,$v) = each %command) {
    $command{$v} = $k;
}

my %action = (
    'c' => 'continue',
    'r' => 'reject',
    'd' => 'discard',
    'a' => 'accept',
    't' => 'tempfail',

    '+' => 'add_recipient',
    '-' => 'remove_recipient',
    'h' => 'add_header',
    'm' => 'change_header',
    'b' => 'replace_body',

    'p' => 'progress',
    'y' => 'set_reply_code',
    'q' => 'quarantine',
    'O' => 'option',
);

sub new {
    my $class = shift;
    unshift(@_, "Socket") if @_ == 1;
    my %arg = @_;

    my $socket  = delete $arg{Socket};
    my $host    = delete $arg{Host};
    my $port    = delete $arg{Port};
    my $path    = delete $arg{Path};
    my $domain  = delete $arg{Domain};
    my $debug   = delete $arg{Debug};
    my $action  = delete $arg{ActionCB};
    my $version = delete $arg{Version} || 2;
    my $flags   = delete $arg{Flags}  || 0xFFFF_FFFF;
    my $pflags  = delete $arg{ProtoFlags} || 0xFFFF_FFFF;
    my $timeout = delete $arg{Timeout};

    if ($^W && %arg) {
	warn "Unknown $class options: " . join(", ", sort keys %arg);
    }

    die "The Host and Path options can not be used together"
	if defined($host) && defined($path);

    if (defined $socket) {
	die "The Socket option can not be combinded with Host or Path"
	    if defined($host) || defined($path);
	require ActiveState::Milter::SockSpec;
        my($scheme, @rest) = ActiveState::Milter::SockSpec::parse($socket);

	if ($scheme eq "unix") {
	    $path = $rest[0];
	}
	else {
	    ($port, $host) = @rest;
	    $domain = $scheme;
	}
    }
   
    my $self;
    if (defined $path) {
	$self = ActiveState::Milter::UNIX->new(Peer => $path,
				       Timeout => $timeout,
				      );
    }
    else {
	$host   ||= "localhost";
	$port   ||= 3366;
	$domain ||= 'inet';

	my $pkg = "ActiveState::Milter::" . uc $domain;
	$self = eval { $pkg->new(PeerAddr => $host,
		                 PeerPort => $port,
		                 Proto    => 'tcp',
		                 Timeout  => $timeout) };
    }
    return undef unless $self;

    $self->debug($debug) if $debug;
    $self->action_cb($action) if $action;

    eval {
	# negotiate options
	$self->send_cmd("O", pack("NNN", $version, $flags, $pflags));

	# Don't want to call here $self->read_resp() until we know
	# that the server speaks the milter protocol.  Normally we
	# would end up with an 'out of memory' error because 4 plain
	# text bytes are likely to be interpreted as a pretty long
	# number.
	my $buf;
	$self->_my_read($buf, 5);
	my($len, $r) = unpack("Na", $buf);
	$len--;
	unless ($r eq "O" && $len == 3 * 4) {
	    die "Not a milter server '$r/$len'";
	}
	$self->_my_read($buf, $len);
	$self->debug_msg("<< $r $buf")
	    if ${*$self}{'mi_debug'};
        ${*$self}{'mi_opts'} = [ unpack("N*", $buf) ];
    };
    return undef if $@;

    return $self;
}

package ActiveState::Milter::Methods;

use Socket  qw(unpack_sockaddr_in unpack_sockaddr_un sockaddr_family
               AF_INET AF_INET6 AF_UNIX);
use Socket6 qw(unpack_sockaddr_in6 inet_ntop);

# some constants
sub SMFIP_NOCONNECT () { 0x00000001 }  # MTA should not send connect info
sub SMFIP_NOHELO    () { 0x00000002 }  # MTA should not send HELO info
sub SMFIP_NOMAIL    () { 0x00000004 }  # MTA should not send MAIL info
sub SMFIP_NORCPT    () { 0x00000008 }  # MTA should not send RCPT info
sub SMFIP_NOBODY    () { 0x00000010 }  # MTA should not send body
sub SMFIP_NOHDRS    () { 0x00000020 }  # MTA should not send headers
sub SMFIP_NOEOH	    () { 0x00000040 }  # MTA should not send EOH

sub version {
    my $self = shift;
    return ${*$self}{'mi_opts'}[0];
}

sub flags {
    my $self = shift;
    return ${*$self}{'mi_opts'}[1];
}

sub proto_flags {
    my $self = shift;
    return ${*$self}{'mi_opts'}[2];
}

sub debug {
    my $self = shift;
    my $old = ${*$self}{'mi_debug'};
    ${*$self}{'mi_debug'} = shift if @_;
    return $old;
}

sub debug_msg {
    my($self, $msg) = @_;

    my $truncated;
    if (length($msg) > 50) {
	substr($msg, 50) = "";
	$truncated++;
    }

    require URI::Escape;
    $msg = URI::Escape::uri_escape($msg, '^\x20-\x7e');
    $msg .= " ..." if $truncated;
    $msg .= "\n" unless $msg =~ /\n\z/;
    warn "ActiveState-Milter: $msg";
}

sub action_cb {
    my $self = shift;
    my $old = ${*$self}{'mi_action_cb'};
    ${*$self}{'mi_action_cb'} = shift if @_;
    return $old;
}

sub mi_connect {
    my($self, $host, $sockaddr) = @_;
    return 'c' if $self->proto_flags & SMFIP_NOCONNECT;
    $host = "host.example.com" unless defined $host;
    my($family, $port, $addr);
    $family = "U"; # unknown
    $port = 0;
    if (defined $sockaddr) {
        my $family_type = sockaddr_family($sockaddr);
        eval {
           if ($family_type == AF_INET) {
               ($port, $addr) = unpack_sockaddr_in($sockaddr);
	       $addr = inet_ntop($family_type, $addr);
               $family = '4';
           }
           elsif ($family_type == AF_INET6) {
               ($port, $addr) = unpack_sockaddr_in6($sockaddr);
	       $addr = inet_ntop($family_type, $addr);
	       $family = '6';
           }
           elsif ($family_type == AF_UNIX) {
               $addr = unpack_sockaddr_un($sockaddr);
	       $family = 'L';
           }
           1;
        } ||
        do {
           $addr = "";
	   $family = 'U'; # unknown
        };
    }
    else {
	$addr = "";
	$family = 'U'; # unknown
    }
    return $self->send_cmd_resp("C", "$host\0" . $family . pack("n", $port) . "$addr\0");
}

sub mi_helo {
    my($self, $msg) = @_;
    return 'c' if $self->proto_flags & SMFIP_NOHELO;
    $msg = "How do you do?" unless defined $msg;
    return $self->send_cmd_resp("H", "$msg\0");
}

sub mi_from {
    my $self = shift;
    return 'c' if $self->proto_flags & SMFIP_NOMAIL;
    return $self->send_cmd_resp("M", _argv(@_));
}

sub mi_to {
    my $self = shift;
    return 'c' if $self->proto_flags & SMFIP_NORCPT;
    return $self->send_cmd_resp("R", _argv(@_));
}

sub mi_header {
    my($self, $k, $v) = @_;
    return 'c' if $self->proto_flags & SMFIP_NOHDRS;
    return $self->send_cmd_resp("L", "$k\0$v\0");
}

sub mi_eoh {
    my $self = shift;
    return 'c' if $self->proto_flags & SMFIP_NOEOH;
    return $self->send_cmd_resp("N");
}

sub mi_body {
    my $self = shift;
    return 'c' if $self->proto_flags & SMFIP_NOBODY;
    return $self->send_cmd_resp("B", join("", @_));
}

sub mi_eom {
    my $self = shift;
    $self->send_cmd("E", ($self->proto_flags & SMFIP_NOBODY) ? "" : join("", @_));

    # read responses until we get a status code,
    # and perform actions
    my $action_cb = $self->action_cb;

    while (1) {
	my($action, $buf) = $self->read_resp;
	if (_is_status($action)) {
	    return ($action, $buf) if wantarray;
	    return $action;
	}

	if ($action_cb) {
	    my @args;
	    if ($action eq "m") { # change_header
		my $index = unpack("N", substr($buf, 0, 4, ""));
		my($k, $v) = split(/\0/, $buf);
		@args = ($k, $index, $v);
	    }
	    elsif ($action eq "h") { # add_header
		@args = split(/\0/, $buf);
	    }
	    else {
		@args = ($buf);
	    }
	    &$action_cb($self, $action{$action} || $action, @args);
	}
	else {
	    my $a = $action{$action} || "action $action";
	    warn "$a\n";
	}
    }
}

sub mi_def_macro {
    my $self = shift;
    my $command = shift;
    $command = $command{$command} || die "Unknown command '$_'"
	unless length($command) == 1;
    return $self->send_cmd("D", join("", $command, map "$_\0", @_));
}

sub mi_abort {
    my $self = shift;
    return $self->send_cmd_resp("A");
}

sub mi_quit {
    my $self = shift;
    $self->send_cmd("Q");
    # no response
    $self->close;
    return;
}

sub send_cmd {
    my($self, $cmd, $data) = @_;
    $data = "" unless defined($data);
    die unless length($cmd) == 1;
    $self->debug_msg(">> $cmd $data") if ${*$self}{'mi_debug'};
    local $\;
    $self->print(pack("N", 1+length($data)) . $cmd . $data) ||
        die "Can't write cmd '$cmd'";
}

sub read_resp {
    my $self = shift;
    my $buf;
    _my_read($self, $buf, 5);
    my($len, $cmd) = unpack("Na", $buf);
    --$len >= 0 || die;
    _my_read($self, $buf, $len);
    $self->debug_msg("<< $cmd $buf")
	if ${*$self}{'mi_debug'};
    return wantarray ? ($cmd, $buf) : $cmd;
}

sub send_cmd_resp {
    my $self = shift;
    $self->send_cmd(@_);
    return $self->read_resp;
}

sub _my_read {
    my $self = shift;
    my $len = $_[1];
    $_[0] = "";
    my $timeout = $self->timeout;
    while ($len) {
	die "read timeout" if $timeout && !_can_read($self, $timeout);
	my $n = sysread($self, $_[0], $len, length($_[0]));
	die "sysread failed: $!" unless defined $n;
	die "unexpected EOF" if $n == 0;
	$len -= $n;
    }
}

sub _argv {
    join("", map "$_\0", @_);
}

sub _is_status {
    $_[0] =~ /[crdaty]/;
}

sub _can_read {
    my($self, $timeout) = @_;
    my $fbits = '';
    vec($fbits, fileno($self), 1) = 1;
    my $nfound = select($fbits, undef, undef, $timeout);
    die "select failed: $!" unless defined $nfound;
    return $nfound > 0;
}

package ActiveState::Milter::INET6;

use base qw(ActiveState::Milter::Methods IO::Socket::INET6);

package ActiveState::Milter::INET;

use base qw(ActiveState::Milter::Methods IO::Socket::INET);

package ActiveState::Milter::UNIX;

use base qw(ActiveState::Milter::Methods IO::Socket::UNIX);

1;

__END__

=head1 NAME

ActiveState::Milter - Milter Client class

=head1 SYNOPSIS

 use ActiveState::Milter;

 my $m = ActiveState::Milter->new(%opts);

 $m->mi_connect("foo.bar.com");
 $m->mi_from("<user\@bar.com>")
 $m->mi_to("<user\@here.com>");
 $m->mi_header("Subject", "Hello!");
 $m->mi_eoh;
 $m->mi_body("How are you?\r\n");
 $m->mi_eom;

 $m->mi_quit;

=head1 DESCRIPTION

The C<ActiveState::Milter> class is used to create milter clients.  The milter
protocol is used by sendmail to talk to mail filters, also called
milters.  In normal cases sendmail acts as the milter client.  The
C<ActiveState::Milter> class is useful because it allows you to test mail
filters without involving sendmail.

The following methods are provided:

=over

=item $m = ActiveState::Milter->new( %args )

This method will try to connect to the filter identified by the given
arguments.  If it cannot connect, then C<undef> is returned (and $@
contains an error message).  Otherwise a C<ActiveState::Milter> object is
returned and methods can be invoked on it to have a conversation with
the filter.

The arguments to this method are passed in as key/value pairs.  The
following arguments are recognized:

   Host        where the filter runs (default 'localhost')
   Port        what port it listens to (default 3366)
   Path        path to unix socket (excludes Host)
   Socket      specify host:port or path using milter socket spec
   Debug       debug mode (default '0')
   ActionCB    see action_cb method below (default is ignore)
   Version     announced version of protocol (default is 2)
   Flags       negotiation bitmask
   ProtoFlags  negotiation bitmask
   Timeout     connect timeout

If the constructor is called with a single argument it is taken to be
socket spec.

=item $m->version

=item $m->flags

=item $m->proto_flags

These three methods can be used to inspect the values of the negotation
parameters that the milter sent back.

=item $m->debug( [$bool] )

This method is used to get/set the current debug value.  If debug is
turned on then a trace of the communcation that takes place will be
printed on STDERR.

=item $m->action_cb( [\&cb] )

This method is used to get/set a callback function that is invoked for
actions that the milter wants sendmail to do.  Actions are only
invoked during the mi_eom call.  The first argument to the action
callback is $m, the second the action name and then there might be
more arguments that depend on the action name.  The action names are:

  add_recipient
  remove_recipient
  add_header
  change_header
  replace_body
  progress
  quarantine

=back

The following request methods are provided.  A call to most of these
return a one-letter status code; 'c', 'r', 'd', 'a', or 't'; for
continue, reject, discard, accept and tempfail respectively.  The
status code 'y' can also be reported instead of 'r' or 't' in cases
the setreply() has been called, and is then (in list context) followed
by a buffer containing the SMTP reply status line.

=over

=item $m->mi_connect( $host, [$sockaddr] )

This method is called to inform the milter that an SMTP connection has
been accepted from host $host.

=item $m->mi_def_macro( $cmd, $k => $v, ... )

This method is called to pass the values of certain symbols to the
milter.  Filters can access these symbols by calling the getsymval()
function in the milter API.

There is no return value reported for this method.

=item $m->mi_helo( $msg )

This method is called to inform the milter about the domain reported
in the HELO command.

=item $m->mi_from( $from, @arg )

This method is called to inform the milter about the "MAIL FROM"
command.  This method must be called at the start of each message.

=item $m->mi_to( $to, @arg )

This method is called to inform the milter about any "RCPT TO"
command.  It can be called multiple times.

=item $m->mi_header( $field_name, $field_value )

Each header of the message is passed on using this method.

=item $m->mi_eoh

And this signals that all the headers have been sent.

=item $m->mi_body( $data, .... )

The body data can be sent in chunks by invoking this method.

=item $m->mi_eom( $data, .... )

And this can pass in the final chunk(s) and also signals that no more
chunks will come.

=item $m->mi_abort

Abort processing of the current message.

=item $m->mi_quit

Ends the connection.  There is no return value reported for this method.

=back

=head1 EXAMPLE

This shows how sendmail typically report a received message to a
milter:

 use ActiveState::Milter;

 my $m = ActiveState::Milter->new(Port => 3366, Debug => shift) || die;

 $m->mi_def_macro("connect", "j", "foo.bar.com");
 $m->mi_connect("xyzzy");

 $m->mi_def_macro("mail_from", "i", "g14LwIgA007058", "{mail_mailer}", "local");
 $m->mi_from("<foo>");
 $m->mi_to("<bar>");

 $m->mi_header("Subject", "Hello!");
 $m->mi_eoh;

 $m->mi_body("How do you do?\r\n");
 $m->mi_eom("\r\n");

 $m->mi_quit;

=head1 BUGS

The timeout mechanism should probably also be applied to writes.

=head1 COPYRIGHT

Copyright (c) 2002 ActiveState Corp.  All rights reserved.

=head1 SEE ALSO

L<PerlMx>, L<sendmail(1)>

http://www.milter.org/

=cut
